/*
 * #%L
 * OW2 Chameleon - Fuchsia Framework
 * %%
 * Copyright (C) 2009 - 2014 OW2 Chameleon
 * %%
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * #L%
 */
/*
    Calimero - A library for KNX network access
    Copyright (C) 2006-2008 B. Malinowsky

    This program is free software; you can redistribute it and/or 
    modify it under the terms of the GNU General Public License 
    as published by the Free Software Foundation; either version 2 
    of the License, or at your option any later version. 
 
    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
    GNU General Public License for more details. 
 
    You should have received a copy of the GNU General Public License 
    along with this program; if not, write to the Free Software 
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. 
 
    Linking this library statically or dynamically with other modules is 
    making a combined work based on this library. Thus, the terms and 
    conditions of the GNU General Public License cover the whole 
    combination. 
 
    As a special exception, the copyright holders of this library give you 
    permission to link this library with independent modules to produce an 
    executable, regardless of the license terms of these independent 
    modules, and to copy and distribute the resulting executable under terms 
    of your choice, provided that you also meet, for each linked independent 
    module, the terms and conditions of the license of that module. An 
    independent module is a module which is not derived from or based on 
    this library. If you modify this library, you may extend this exception 
    to your version of the library, but you are not obligated to do so. If 
    you do not wish to do so, delete this exception statement from your 
    version. 
*/

package tuwien.auto.calimero.link.medium;

import tuwien.auto.calimero.exception.KNXFormatException;

/**
 * Factory for raw frames on medium.
 * <p>
 * Supports creation of raw frames out of byte arrays for now.
 * 
 * @author B. Malinowsky
 */
public final class RawFrameFactory
{
	private RawFrameFactory()
	{}

	/**
	 * Creates a raw frame out of a byte array for the specified communication medium.
	 * <p>
	 * This method just invokes one of the other medium type specific creation methods
	 * according the given medium type.
	 * 
	 * @param mediumType KNX communication medium, one of the media types declared in
	 *        {@link KNXMediumSettings}
	 * @param data byte array containing the raw frame structure
	 * @param offset start offset of frame structure in <code>data</code>, 0 &lt;=
	 *        offset &lt; <code>data.length</code>
	 * @return the created raw frame for the specified medium
	 * @throws KNXFormatException on unknown/not supported KNX medium or no valid frame
	 *         structure
	 */
	public static RawFrame create(int mediumType, byte[] data, int offset)
		throws KNXFormatException
	{
		switch (mediumType) {
		case KNXMediumSettings.MEDIUM_TP0:
			throw new KNXFormatException("TP0 raw frame not supported yet");
		case KNXMediumSettings.MEDIUM_TP1:
			return createTP1(data, offset);
		case KNXMediumSettings.MEDIUM_PL110:
			return createPL110(data, offset);
		case KNXMediumSettings.MEDIUM_PL132:
			return createPL132(data, offset);
		case KNXMediumSettings.MEDIUM_RF:
			throw new KNXFormatException("RF raw frame not supported yet");
		default:
			throw new KNXFormatException("unknown KNX medium for raw frame", mediumType);
		}
	}

	/**
	 * Creates a raw frame out of a byte array for the TP1 communication medium.
	 * <p>
	 * 
	 * @param data byte array containing the TP1 raw frame structure
	 * @param offset start offset of frame structure in <code>data</code>, 0 &lt;=
	 *        offset &lt; <code>data.length</code>
	 * @return the created TP1 raw frame
	 * @throws KNXFormatException on no valid frame structure
	 */
	public static RawFrame createTP1(byte[] data, int offset) throws KNXFormatException
	{
		final int ctrl = data[offset] & 0xff;
		// parse control field and check if valid
		if ((ctrl & 0x10) == 0x10) {
			if ((ctrl & 0x40) == 0x00)
				return new TP1LData(data, offset);
			else if (ctrl == 0xF0)
				return new TP1LPollData(data, offset);
			throw new KNXFormatException("invalid raw frame control field", ctrl);
		}
		return new TP1Ack(data, offset);
	}

	/**
	 * Creates a raw frame out of a byte array for the PL110 communication medium.
	 * <p>
	 * 
	 * @param data byte array containing the PL110 raw frame structure
	 * @param offset start offset of frame structure in <code>data</code>, 0 &lt;=
	 *        offset &lt; <code>data.length</code>
	 * @return the created PL110 raw frame
	 * @throws KNXFormatException on no valid frame structure
	 */
	public static RawFrame createPL110(byte[] data, int offset) throws KNXFormatException
	{
		if ((data[0] & 0x10) == 0x10)
			return new PL110LData(data, offset);
		return new PL110Ack(data, offset);
	}

	/**
	 * Creates a raw frame out of a byte array for the PL132 communication medium.
	 * <p>
	 * 
	 * @param data byte array containing the PL132 raw frame structure
	 * @param offset start offset of frame structure in <code>data</code>, 0 &lt;=
	 *        offset &lt; <code>data.length</code>
	 * @return the created PL132 raw frame
	 * @throws KNXFormatException on no valid frame structure
	 */
	public static RawFrame createPL132(byte[] data, int offset) throws KNXFormatException
	{
		if (data.length - offset == 2)
			return new PL132Ack(data, offset);
		return new PL132LData(data, offset);
	}
}
